class Bidict:
    ''' The simplest class that implements a bijective dict '''

    def __init__(self, init:dict = {}):
        _data = init
        for key in init.keys():
            _data[init[key]] = key
    
    def __setitem__(self, key, value):
        self._data[key] = value
        self._data[value] = key
    
    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[self._data[key]]
        del self._data[key]

    def __contains__(self, key):
        return key in self._data



import socket
import threading
from time import sleep

class Client:
    ''' A class that implements the logic of connecting, authorizing, and exchanging messages with the server '''
    host = None
    port = None
    connected = False
    def __init__(self, username, chat_id):
        self.socket = socket.socket()
        self.username = username
        self.chat_id = chat_id

    def connect(self, host, port):
        ''' Sends authorization information to the server and responds to the returned code. Starts listen thread. '''
        self.host = host
        self.port = port
        try:
            self.socket.connect((host,port))
            self.send(f'{self.username}_{self.chat_id}')
            answer = self.receive()
            assert answer == '200'
            self.connected = True
            listen_thread = threading.Thread(target=self.listen, daemon=True)
            listen_thread.start()
        except AssertionError:
            if answer == '403':
                print('The user with such name already connect')
                self.socket.close()
            elif answer == '401':
                print('Authorization error')
                self.socket.close()
        except:
            print('Connection error')
            self.socket.close()
        else:
            print('Connected!')

    def shell(self):
        ''' A shell for reading messages from the console and sending them to the server '''
        try:
            while self.connected:
                sleep(0.2)
                message = input('>> ')
                if not message:
                    print('I can`t send empty message')
                elif message == '/exit':
                    print('Exit!')
                    break
                elif message == '/down':
                    self.socket.shutdown(socket.SHUT_RDWR)
                    break
                else:
                    self.send(message)
        except KeyboardInterrupt:
            print(' - you tab CTRL-C')
        finally:
            print('Closing...')
            self.close()
    
    def listen(self):
        ''' Listens to the socket and outputs messages '''
        try:
            while self.connected:
                data = self.socket.recv(1024)
                if len(data) > 0:
                    message = data.decode("utf-8")
                    if message=='/exit':
                        self.connected = False
                    elif message.startswith(self.username):
                        print(message)
                    else:
                        print('\n',message)
                else:
                    self.connected = False
        finally:
            print('\nStop listening...')
            self.close()
            exit()

    def send(self, message: str):
        ''' Encodes and send message'''
        self.socket.send(message.encode('utf-8'))
    
    def receive(self):
        ''' Receives and returns message '''
        data = self.socket.recv(1024)
        return data.decode('utf-8')
    

    def close(self):
        ''' Close connection and stops listen and shell loops '''
        self.connected = False
        self.socket.close()





# -*- coding: utf-8 -*

import threading
import logging
import logging.handlers
from abc import ABC, abstractmethod
from socket import socket
import time
from typing import MutableSet


class LoopThread(ABC, threading.Thread):
    ''' LoopThread runs on a separate thread and includes an infinite loop that executes `main` and calls `final` upon completion. '''
    active = True
    def run(self):
        try:
            while self.active:
                self.main()
        finally:
            self.final()

    @abstractmethod
    def main(self):
        ''' The function to be called in the loop until the loop is stopped '''
        pass

    @abstractmethod
    def final(self):
        ''' Сalled after the end of the loop '''
        pass

    def stop(self):
        ''' Stops the loop '''
        self.active = False

class Chat:
    pass

class Connection(LoopThread):
    ''' Thread of user connection to server '''
    chat:Chat = None
    def __init__(connection, sock: socket, username:str, chat_id:str):
        super().__init__(daemon=True)
        connection.sock = sock
        connection.username = username
        connection.chat_id = chat_id
    
    def receive(connection) -> str:
        ''' Receives and decodes data '''
        try:
            return connection.sock.recv(1024).decode('utf-8')
        except OSError:
            return ''

    def send(connection, message: str):
        ''' Encode and send message to client '''
        connection.sock.send(message.encode('utf-8'))

class Server:
    ''' Server for multithreading TCP connections '''
    def __init__(self, host='localhost', port=8888):
        self.host = host
        self.port = port
        self.socket:socket = socket()
        self.socket.bind((host, port))
        self.connections = {}

        self.logger = logging.getLogger('TCP_server')
        self.logger.setLevel('INFO')
        self.logger.addHandler(logging.handlers.SysLogHandler(address='/dev/log'))

        class Connection_(Connection):
            ''' Фn internal connection class that defines the logic for a specific server '''
            def main(connection):
                data = connection.receive()
                if len(data) > 0:
                    message = f'{connection.username}: {data}'
                    self.logger.info(message)
                    connection.chat.send(connection.username, data)
                else:
                    connection.stop()
                pass

            def final(connection):
                connection.send('/exit')
                connection.sock.close()
                del self.connections[connection.username]
                print(connection.username, 'disconnected')

        self.Connection = Connection_ 
        self.chats:dict[str, Connection_] = {}


    def receive(self, n:int = 1024, encoding:str = 'utf-8') -> str:
        ''' Receives and decrypts data from a socket '''
        return self.socket.recv(n).decode(encoding)

    def auth(self ,sock: socket, addr: str) -> Connection:
        ''' Attempts to authorize a user on this server. Sends one of three possible response codes to the client:
            * 200 - authorization is successful
            * 400 - invalid data
            * 403 - a user with this nickname is already logged in \n  
            Returns a connection object.'''
        data = sock.recv(1024).decode('utf-8')
        try:
            username, chat_id  = data.split('_')
        except ValueError:
            sock.send(b'400')
            sock.close()
            print('Authentificated error with', addr)
            return None
        if username and chat_id and username not in self.connections.keys():
            print("Connected to", addr, 'as', username)
            sock.send(b'200')
            connection = self.Connection(sock = sock, username = username, chat_id = chat_id)
            connection.start()
            self.connections[username] = connection
            print('Num of connections:', len(self.connections))
            return connection

        elif username in self.connections.keys():
            sock.send(b'403')
            sock.close()
            return None
        else:
            sock.send(b'400')
            sock.close()
            print('Authentificated error with', addr)
            return None

    def join_chat(self, chat_id, username):
        ''' Adds a user to the chat with the chat_id identifier. If there is no such chat, then creates it '''
        user = self.connections[username]
        if chat_id in self.chats.keys():
            self.chats[chat_id].join_user(user)
            user.chat = self.chats[chat_id]
        else:
            chat = Chat(chat_id)
            user.chat = chat
            self.chats[chat_id] = chat
            chat.start()
            chat.join_user(user)

    def start(self):
        ''' Starts mainloop of server '''
        print('Start server')
        self.socket.listen(5)
        try:
            while True:
                sock, addr = self.socket.accept()
                conn = self.auth(sock, addr)
                if conn:
                    self.join_chat(conn.chat_id, conn.username)
        except KeyboardInterrupt:
            print(' - you tab CTRL-C')
        finally:
            print('Stop server')
            for connection in self.connections.values():
                connection.stop()
            self.socket.close()


class Chat(LoopThread):
    ''' Class for implementing the logic for sending messages to groups '''
    active = True
    id: str = None
    def __init__(self, id:str):
        super(Chat, self).__init__(daemon=True)
        self.id = id
        self.members:MutableSet[Connection] = set()
        self.lock = threading.Lock()
    
    def join_user(self, conn: Connection):
        ''' Joins user to member set '''
        self.lock.acquire()
        self.members.add(conn)
        self.lock.release()
        print(self.id, '-', self.members)

    def send(self, sender, text):
        ''' Sends message to each member '''
        print(text,1213)
        self.lock.acquire()
        for member in self.members:
            member.send(f'{sender} - {text}')
        self.lock.release()

    def main(self):
        
        # for conn_sender in self.members:
        #     data = conn_sender.receive()
        #     if len(data):
        #         for conn_receiver in self.members:
        #             conn_receiver.send(f'{conn_sender.username}-{data}')
        time.sleep(1)

    def final(self):
        ''' Close all connections of this chat '''
        print('Chat closed')
        for conn in self.members:
            conn.stop()            



#!/bin/python3
import argparse
from .server import Server
from .client import Client
from time import sleep

parser = argparse.ArgumentParser(description='')
parser.add_argument('side', type=str, help='client or server')
parser.add_argument('-a', '--adress', help='IP of your server', default='localhost')
parser.add_argument('-p', '--port', help='Port of your server', default='8888')
parser.add_argument('-u', '--username', help='Your name in the chats')
parser.add_argument('-c', '--chat', help='Id of an existing or new chat')
args = parser.parse_args()

if args.side == 'server':
    serv = Server(args.adress, int(args.port))
    serv.start()
elif args.side == 'client':
    cl = Client(args.username, args.chat)
    cl.connect(args.adress, int(args.port))
    if cl.connected:
        cl.shell()
else:
    print('ERROR ARGUMENT')


from setuptools import setup, find_packages

with open("README.md") as file:
    read_me_description = file.read()


setup(name='cn_tcp_server',
      version='0.0.2',
      description='Simple server and client for chatting',
      long_description=read_me_description,
      long_description_content_type="text/markdown",
      url='https://github.com/cl1ckname/tcp_chat',
      packages=find_packages(),
      classifiers=[
        "Programming Language :: Python :: 3.6",
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Topic :: Communications :: Chat",
      ],
      python_requires=">=3.6",
      author_email='20002mc@gmail.com',
      zip_safe=False)  